@@ -1,4 +1,7 @@
-/* Copyright (c) 1993 - 1998, Joseph Arceneaux.  All rights reserved.
+/*
+   Copyright 1999-2002,2010, Thomas E. Dickey
+
+   Copyright (c) 1993 - 1998, Joseph Arceneaux.  All rights reserved.
 
    This file is subject to the terms of the GNU General Public License as
    published by the Free Software Foundation.  A copy of this license is
@@ -16,23 +19,18 @@
 #include "sys.h"
 #include "indent.h"
 
-#include <string.h>
-
 /* Check the limits of the comment buffer, and expand as neccessary. */
 
 #define CHECK_COM_SIZE \
 	if (e_com >= l_com) \
           { \
-	    int nsize = l_com-s_com+400; \
+	    size_t nsize = (size_t) (l_com - s_com + 400); \
 	    combuf = (char *) xrealloc (combuf, nsize); \
-	    e_com = combuf + (e_com-s_com) + 1; \
+	    e_com = combuf + (e_com - s_com) + 1; \
 	    l_com = combuf + nsize - 5; \
 	    s_com = combuf + 1; \
 	  }
 
-/* The number of comments handled so far. */
-int out_coms;
-
 /* Output a comment.  `buf_ptr' is pointing to the character after
    the beginning comment delimiter when this is called.  This handles
    both C and C++ comments.
@@ -75,31 +73,31 @@ int out_coms;
    */
 
 void
-print_comment ()
+print_comment (void)
 {
   register int column, format;
   enum codes comment_type;
 
   int start_column, found_column;
   int first_comment_line, right_margin;
-  int boxed_comment, stars, blankline_delims, paragraph_break,
-      merge_blank_comment_lines;
+  int boxed_comment, stars, blankline_delims, paragraph_break, merge_blank_comment_lines;
 
   int two_contiguous_comments = 0;
   int save_length = 0;
   char *save_ptr = 0;
   char *text_on_line = 0;
   char *line_break_ptr = 0;
-  char *start_delim, *end_delim;
+  const char *start_delim;
 
-  char *line_preamble;
-  int line_preamble_length, visible_preamble;
+  const char *line_preamble;
+  int line_preamble_length;
+  int visible_preamble;
 
   /* Increment the parser stack, as we will store some things
      there for dump_line to use. */
   inc_pstack ();
 
-  /* Have to do it this way because this piece of shit program doesn't
+  /* Have to do it this way because this program doesn't
      always place the last token code on the stack. */
   if (*(token + 1) == '/')
     comment_type = cplus_comment;
@@ -122,7 +120,7 @@ print_comment ()
     }
   else if (*buf_ptr == '*' || *buf_ptr == '-'
 	   || *buf_ptr == '=' || *buf_ptr == '_'
-	   || (parser_state_tos->col_1 && ! format_col1_comments))
+	   || (parser_state_tos->col_1 && !format_col1_comments))
     /* Boxed comment.  This block of code will return. */
     {
       int comment_lines_count = 1;
@@ -139,20 +137,23 @@ print_comment ()
       while (1)
 	{
 	  do
-	    *e_com++ = *buf_ptr++;
-	  while (*buf_ptr != '*' && buf_ptr < buf_end);
+	    {
+	      *e_com++ = *buf_ptr++;
+	      CHECK_COM_SIZE;
+	    }
+	  while (*buf_ptr != '*' && !at_buffer_end (buf_ptr));
 
 	  CHECK_COM_SIZE;
 
 	  /* We have reached the end of the comment, and it's all on
-             this line. */
+	     this line. */
 	  if (*buf_ptr == '*' && *(buf_ptr + 1) == '/')
 	    {
-	      if (buf_ptr == buf_end)
+	      if (at_buffer_end (buf_ptr))
 		fill_buffer ();
 
 	      buf_ptr += 2;
-	      if (buf_ptr == buf_end)
+	      if (at_buffer_end (buf_ptr))
 		fill_buffer ();
 
 	      *e_com++ = '*';
@@ -161,10 +162,10 @@ print_comment ()
 	      parser_state_tos->tos--;
 
 	      /* If this is the only line of a boxed comment, it may
-		 be after some other text (e.g., #if foo <comment>),
-		 in which case we want to specify the correct column.
-		 In the other cases, the leading spaces account for
-		 the columns and we start it in column 1. */
+	         be after some other text (e.g., #if foo <comment>),
+	         in which case we want to specify the correct column.
+	         In the other cases, the leading spaces account for
+	         the columns and we start it in column 1. */
 
 	      if (comment_lines_count > 1)
 		parser_state_tos->com_col = 1;
@@ -174,7 +175,7 @@ print_comment ()
 	    }
 
 	  /* End of the line, or end of file. */
-	  if (buf_ptr == buf_end)
+	  if (at_buffer_end (buf_ptr))
 	    {
 	      if (*(buf_ptr - 1) == EOL)
 		{
@@ -198,7 +199,6 @@ print_comment ()
   else
     {
       start_delim = "/*";
-      end_delim = "*/";
       line_preamble = 0;
       line_preamble_length = 0;
       visible_preamble = 0;
@@ -219,13 +219,13 @@ print_comment ()
   if ((s_lab == e_lab) && (s_code == e_code))
     /* First handle comments which begin the line. */
     {
-      if (parser_state_tos->col_1 && ! format_col1_comments)
+      if (parser_state_tos->col_1 && !format_col1_comments)
 	{
 	  format = format_col1_comments;
 	  start_column = 1;
 	}
       else if (s_com != e_com)
-	/* The fool has a line consisting of two contiguous comments.
+	/* We have a line consisting of two contiguous comments.
 	   In this case, we don't try too hard, 'cause nothing will
 	   look good. */
 	{
@@ -267,7 +267,7 @@ print_comment ()
 	target = com_ind;
 
       /* Now determine if the code on the line is short enough
-	 to allow the comment to begin where it should. */
+         to allow the comment to begin where it should. */
       if (s_code != e_code)
 	start_column = count_columns (compute_code_target (),
 				      s_code, NULL_CHAR);
@@ -295,7 +295,7 @@ print_comment ()
       format = format_comments;
     }
 
-  if (! line_preamble)
+  if (!line_preamble)
     {
       line_preamble_length = 3;
       if (stars)
@@ -317,6 +317,19 @@ print_comment ()
 
   /* Output the beginning comment delimiter.  They are both two
      characters long. */
+  CHECK_COM_SIZE;
+  if (e_com != s_com)
+    {
+      if (blankline_delims)
+	{
+	  *e_com = '\0';
+	  dump_line ();
+	}
+      else
+	{
+	  *e_com++ = ' ';
+	}
+    }
   *e_com++ = *start_delim;
   *e_com++ = *(start_delim + 1);
   column = start_column + 2;
@@ -330,35 +343,32 @@ print_comment ()
       dump_line ();
 
       /* Check if the delimiter was already on a line by itself,
-	 and skip whitespace if formating. */
-      while (*p == ' ' || *p == TAB)
+         and skip whitespace if formating. */
+      while (isblank (*p))
 	p++;
       if (*p == EOL)
 	buf_ptr = p + 1;
       else if (format)
 	buf_ptr = p;
-      if (buf_ptr >= buf_end)
+      if (at_buffer_end (buf_ptr))
 	fill_buffer ();
 
-      column = start_column;
       goto begin_line;
     }
   else if (format)
     {
       *e_com++ = ' ';
       column = start_column + 3;
-      while (*buf_ptr == ' ' || *buf_ptr == TAB)
-	if (++buf_ptr >= buf_end)
+      while (isblank (*buf_ptr))
+	if (at_buffer_end (++buf_ptr))
 	  fill_buffer ();
     }
 
   /* Iterate through the lines of the comment */
-  while (! had_eof)
+  while (!had_eof)
     {
-      int whitespace_character;	/* Mark if we've just seen whitespace or not */
-
       /* Iterate through the characters on one line */
-      while (! had_eof)
+      while (!had_eof)
 	{
 	  CHECK_COM_SIZE;
 
@@ -366,11 +376,9 @@ print_comment ()
 	    {
 	    case ' ':
 	    case TAB:
-	      whitespace_character = 1;
-
 	      /* If formatting, and previous break marker is
 	         nonexistant, or before text on line, reset
-		 it to here. */
+	         it to here. */
 	      if (format && line_break_ptr < text_on_line)
 		line_break_ptr = e_com;
 
@@ -385,8 +393,8 @@ print_comment ()
 		     based on the column we found the comment in, not
 		     the one we're printing in. */
 		  int tab_width
-		    = (tabsize - ((column + found_column - start_column - 1)
-				  % tabsize));
+		  = (tabsize - ((column + found_column - start_column - 1)
+				% tabsize));
 		  column += tab_width;
 		  while (tab_width--)
 		    *e_com++ = ' ';
@@ -395,8 +403,6 @@ print_comment ()
 
 
 	    case EOL:
-	      whitespace_character = 1;
-
 	      /* We may be at the end of a C++ comment */
 	      if (comment_type == cplus_comment)
 		{
@@ -414,7 +420,7 @@ print_comment ()
 		     end an input line, so check here if we need to
 		     get the next line. */
 		  buf_ptr++;
-		  if (buf_ptr >= buf_end)
+		  if (at_buffer_end (buf_ptr))
 		    fill_buffer ();
 
 		  /* If there are any spaces between the text and this
@@ -425,10 +431,10 @@ print_comment ()
 
 		  /* If this is "\n\n", or "\n<whitespace>\n",
 		     it's a paragraph break. */
-		  while (*buf_ptr == TAB || *buf_ptr == ' ')
-		    if (buf_ptr++ >= buf_end)
+		  while (isblank (*buf_ptr))
+		    if (at_buffer_end (buf_ptr++))
 		      fill_buffer ();
-		  if (*buf_ptr == EOL || ! text_on_line)
+		  if (*buf_ptr == EOL || !text_on_line)
 		    {
 		      paragraph_break = 1;
 		      goto end_line;
@@ -445,7 +451,6 @@ print_comment ()
 		/* We are printing this line "as is", so output it
 		   and continue on to the next line. */
 		goto end_line;
-	      break;
 
 	    case '*':
 	      /* Check if we've reached the end of the comment. */
@@ -455,8 +460,8 @@ print_comment ()
 		    {
 		      /* If it's not a boxed comment, put some whitespace
 		         before the ending delimiter.  Otherwise, simply
-			 insert the delimiter. */
-		      if (! boxed_comment)
+		         insert the delimiter. */
+		      if (!boxed_comment)
 			{
 			  if (text_on_line)
 			    {
@@ -468,12 +473,12 @@ print_comment ()
 				}
 			      else
 				/* Insert space before closing delim */
-				if (*(e_com - 1) != ' '&& *(e_com - 1) != TAB)
-				  *e_com++ = ' ';
+			      if (!isblank (*(e_com - 1)))
+				*e_com++ = ' ';
 			    }
 			  /* If no text on line, then line is completely empty
-			      or starts with preamble, or is beginning of
-			      comment and starts with beginning delimiter. */
+			     or starts with preamble, or is beginning of
+			     comment and starts with beginning delimiter. */
 			  else if (s_com == e_com
 				   || *s_com != '/')
 			    {
@@ -483,10 +488,10 @@ print_comment ()
 			  else
 			    /* This is case of first comment line.  Test
 			       with:
-			         if (first_comment_line != com_lines)
-				   abort (); */
-			    if (*(e_com - 1) != ' ' && *(e_com - 1) != TAB)
-			      *e_com++ = ' ';
+			       if (first_comment_line != com_lines)
+			       abort (); */
+			  if (!isblank (*(e_com - 1)))
+			    *e_com++ = ' ';
 			}
 
 		      /* Now insert the ending delimiter */
@@ -495,14 +500,14 @@ print_comment ()
 		      *e_com = '\0';
 
 		      /* Skip any whitespace following the comment.  If
-			 there is only whitespace after it, print the line.
+		         there is only whitespace after it, print the line.
 
-			 NOTE:  We're not printing the line: TRY IT! */
+		         NOTE:  We're not printing the line: TRY IT! */
 		      buf_ptr += 2;
-		      while (*buf_ptr == ' ' || *buf_ptr == TAB)
+		      while (isblank (*buf_ptr))
 			buf_ptr++;
-		      if (buf_ptr >= buf_end)
-			  fill_buffer ();
+		      if (at_buffer_end (buf_ptr))
+			fill_buffer ();
 
 		      parser_state_tos->tos--;
 		      parser_state_tos->com_col
@@ -531,12 +536,12 @@ print_comment ()
 		      break;
 		    }
 		}
+	      /* FALLTHRU */
 	      /* If it was not the end of the comment, drop through
 	         and insert the star on the line. */
 
 	    default:
 	      /* Some textual character. */
-	      whitespace_character = 0;
 	      text_on_line = e_com;
 	      *e_com++ = *buf_ptr;
 	      column++;
@@ -554,7 +559,7 @@ print_comment ()
 		{
 		  *line_break_ptr = '\0';
 		  save_ptr = line_break_ptr + 1;
-		  save_length = e_com - save_ptr;
+		  save_length = (int) (e_com - save_ptr);
 		  e_com = line_break_ptr;
 
 		  /* If we had to go past `right_margin' to print stuff out,
@@ -568,7 +573,7 @@ print_comment ()
 		{
 		  if (comment_type == cplus_comment)
 		    {
-		      while (*buf_ptr == TAB || *buf_ptr == ' ')
+		      while (isblank (*buf_ptr))
 			buf_ptr++;
 		      buf_ptr--;
 		      if (*buf_ptr == EOL)
@@ -576,7 +581,7 @@ print_comment ()
 		    }
 		  else
 		    {
-		      while (*buf_ptr == TAB || *buf_ptr == ' '
+		      while (isblank (*buf_ptr)
 			     || *buf_ptr == EOL)
 			buf_ptr++;
 
@@ -589,28 +594,28 @@ print_comment ()
 	    }
 
 	  buf_ptr++;
-	  if (buf_ptr == buf_end)
+	  if (at_buffer_end (buf_ptr))
 	    fill_buffer ();
 	}
 
 
     end_line:
       /* Compress pure whitespace lines into newlines. */
-      if (! text_on_line
-	  && ! visible_preamble
-	  && ! (first_comment_line == com_lines))
+      if (!text_on_line
+	  && !visible_preamble
+	  && !(first_comment_line == com_lines))
 	e_com = s_com;
       *e_com = '\0';
       dump_line ();
 
       /* If formatting (paragraph_break is only used for formatted
-	 comments) and user wants blank lines merged, kill all white
-	 space after the "\n\n" indicating a paragraph break. */
+         comments) and user wants blank lines merged, kill all white
+         space after the "\n\n" indicating a paragraph break. */
       if (paragraph_break)
 	{
 	  if (merge_blank_comment_lines)
-	    while (*buf_ptr == EOL || *buf_ptr == ' ' || *buf_ptr == TAB)
-	      if (++buf_ptr >= buf_end)
+	    while (*buf_ptr == EOL || isblank (*buf_ptr))
+	      if (at_buffer_end (++buf_ptr))
 		fill_buffer ();
 	  paragraph_break = 0;
 	}
@@ -619,7 +624,7 @@ print_comment ()
 	  /* If it was a paragraph break (`if' clause), we scanned ahead
 	     one character.  So, here in the `else' clause, advance buf_ptr. */
 	  buf_ptr++;
-	  if (buf_ptr >= buf_end)
+	  if (at_buffer_end (buf_ptr))
 	    fill_buffer ();
 	}
 
@@ -628,12 +633,12 @@ print_comment ()
 	break;
 
       /* Indent the line properly.  If it's a boxed comment, align with
-	 the '*' in the beginning slash-star and start inserting there.
-	 Otherwise, insert blanks for alignment, or a star if the
-	 user specified -sc. */
+         the '*' in the beginning slash-star and start inserting there.
+         Otherwise, insert blanks for alignment, or a star if the
+         user specified -sc. */
       if (line_preamble)
 	{
-	  (void) memcpy (e_com, line_preamble, line_preamble_length);
+	  (void) memcpy (e_com, line_preamble, (size_t) line_preamble_length);
 	  e_com += line_preamble_length;
 	  column = start_column + line_preamble_length;
 	}
@@ -643,32 +648,46 @@ print_comment ()
 
       /* If we have broken the line before the end for formatting,
          copy the text after the break onto the beginning of this
-	 new comment line. */
+         new comment line. */
       if (save_ptr)
 	{
-	  while ((*save_ptr == ' ' || *save_ptr == TAB) && save_length)
+	  while (isblank (*save_ptr) && save_length)
 	    {
 	      save_ptr++;
 	      save_length--;
 	    }
-	  (void) memcpy (e_com, save_ptr, save_length);
+	  (void) memcpy (e_com, save_ptr, (size_t) save_length);
 	  text_on_line = e_com;
 	  e_com += save_length;
 	  /* We only break if formatting, in which cases there
-	     are no tabs, only spaces.*/
+	     are no tabs, only spaces. */
 	  column += save_length;
 	  save_ptr = 0;
 	  save_length = 0;
 	}
       else
 	{
-	  while (*buf_ptr == ' ' || *buf_ptr == TAB)
-	    if (++buf_ptr >= buf_end)
+	  while (isblank (*buf_ptr))
+	    if (at_buffer_end (++buf_ptr))
 	      fill_buffer ();
 	  text_on_line = 0;
 	}
     }
 
+  /*
+   * Recover from special case of a comment on the last line of a file which
+   * has no trailing newline.
+   */
+  if (had_eof && !at_buffer_end (buf_ptr))
+    {
+      while (!at_buffer_end (buf_ptr))
+	{
+	  CHECK_COM_SIZE;
+	  *e_com++ = *buf_ptr++;
+	}
+      --e_com;
+    }
+
   parser_state_tos->tos--;
   parser_state_tos->com_col = (two_contiguous_comments ? 1 : start_column);
   parser_state_tos->box_com = boxed_comment;
