@@ -1,4 +1,7 @@
-/* Copyright (c) 1994, Joseph Arceneaux.  All rights reserved.
+/*
+   Copyright 1999-2002,2010, Thomas E. Dickey
+
+   Copyright (c) 1994, Joseph Arceneaux.  All rights reserved.
 
    Copyright (c) 1992, Free Software Foundation, Inc.  All rights reserved.
 
@@ -22,20 +25,12 @@
 
 #include "sys.h"
 #include "indent.h"
-#include <ctype.h>
-#include <stdlib.h>
-
-#if defined (HAVE_UNISTD_H)
-#include <unistd.h>
-#endif
-#include <string.h>
-
 
 #ifdef VMS
 #   include <file.h>
 #   include <types.h>
 #   include <stat.h>
-#else  /* not VMS */
+#else /* not VMS */
 
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -49,12 +44,9 @@
 
 #endif /* not VMS */
 
-extern void free ();
-
 /* number of levels a label is placed to left of code */
 #define LABEL_OFFSET 2
 
-
 /* Stuff that needs to be shared with the rest of indent. Documented in
    indent.h.  */
 char *in_prog;
@@ -65,16 +57,18 @@ FILE *output;
 char *buf_ptr;
 char *buf_end;
 char *buf_break;
-int  break_line;
+int break_line;
 int had_eof;
 int out_lines;
 int com_lines;
+int indent_eqls;
+int indent_eqls_1st;
 
 int suppress_blanklines = 0;
-static int comment_open;
+static int not_first_line;
 
 int paren_target;
-
+
 #ifdef VMS
 /* Folks say VMS requires its own read routine.  Then again, some folks
    say it doesn't.  Different folks have also sent me conflicting versions
@@ -86,39 +80,93 @@ int paren_target;
 int
 vms_read (int file_desc, char *buffer, int nbytes)
 {
-    char *bufp;
-    int nread, nleft;
+  char *bufp;
+  int nread, nleft;
 
-    bufp  = buffer;
-    nread = 0;
-    nleft = nbytes;
+  bufp = buffer;
+  nread = 0;
+  nleft = nbytes;
 
-    nread = read (file_desc, bufp, nleft);
-    while (nread > 0)
-      {
-        bufp += nread;
-        nleft -= nread;
-        if (nleft < 0)
-	  fatal ("Internal buffering error");
-	nread = read (file_desc, bufp, nleft);
-      }
+  nread = read (file_desc, bufp, nleft);
+  while (nread > 0)
+    {
+      bufp += nread;
+      nleft -= nread;
+      if (nleft < 0)
+	fatal ("Internal buffering error");
+      nread = read (file_desc, bufp, nleft);
+    }
 
-    return nbytes - nleft;
+  return nbytes - nleft;
 }
 #endif /* VMS */
-
-INLINE int
-count_columns (column, bp, stop_char)
-     int column;
-     char *bp;
-     char stop_char;
+
+static void
+pass_char (int ch)
+{
+  if (ch == '\n')
+    ++out_line_no;
+  putc (ch, output);
+}
+
+static void
+pass_text (const char *s)
+{
+  while (*s != 0)
+    pass_char (*s++);
+}
+
+static char *
+pass_line (char *p)
+{
+  while (*p != '\0' && *p != EOL)
+    pass_char (*p++);
+  if (*p == EOL)
+    {
+      pass_char ('\n');
+      ++in_line_no;
+      ++p;
+      in_prog_pos = p;
+      cur_line = p;
+    }
+  else if (*p == '\0'
+	   && ((unsigned int) (p - current_input->data) == current_input->size))
+    {
+      buf_ptr = buf_end = in_prog_pos = p;
+      had_eof = true;
+      p = 0;
+    }
+  return p;
+}
+
+static void
+pass_n_text (const char *s, int n)
+{
+  while (n-- > 0)
+    pass_char (*s++);
+}
+
+static void
+reset_output (void)
+{
+  reset_parser ();
+  not_first_line = 0;
+  n_real_blanklines = 0;
+  suppress_blanklines = 1;
+}
+
+int
+count_columns (
+		int column,
+		char *bp,
+		int stop_char)
 {
   while (*bp != stop_char && *bp != NULL_CHAR)
     {
       switch (*bp++)
 	{
 	case EOL:
-	case 014:		/* form feed */
+	case FORM_FEED:	/* form feed */
 	  column = 1;
 	  break;
 	case TAB:
@@ -136,31 +184,15 @@ count_columns (column, bp, stop_char)
   return column;
 }
 
-/* Return the column we are at in the input line. */
-
-INLINE int
-current_column ()
+static int
+this_column (char *from, char *to, int column)
 {
-  char *p;
-  int column;
-
-  if (buf_ptr >= save_com.ptr && buf_ptr <= save_com.ptr + save_com.len)
+  while (from < to)
     {
-      p = save_com.ptr;
-      column = save_com.column;
-    }
-  else
-    {
-      p = cur_line;
-      column = 1;
-    }
-
-  while (p < buf_ptr)
-    {
-      switch (*p)
+      switch (*from)
 	{
 	case EOL:
-	case 014:		/* form feed */
+	case FORM_FEED:	/* form feed */
 	  column = 1;
 	  break;
 
@@ -177,48 +209,82 @@ current_column ()
 	  break;
 	}
 
-      p++;
+      from++;
+    }
+  return column;
+}
+
+/* Return the column we are at in the input line. */
+
+int
+current_column (void)
+{
+  char *p;
+  int column;
+
+  if (buf_ptr >= save_com.ptr && buf_ptr <= save_com.ptr + save_com.len)
+    {
+      p = save_com.ptr;
+      column = save_com.column;
+    }
+  else
+    {
+      p = cur_line;
+      column = 1;
     }
 
-  return column;
+  return this_column (p, buf_ptr, column);
 }
 
 /* Fill the output line with whitespace up to TARGET_COLUMN, given that
    the line is currently in column CURRENT_COLUMN.  Returns the ending
    column. */
 
-INLINE int
-pad_output (current_column, target_column)
-  int current_column;
-  int target_column;
+static int
+pad_output (
+	     int my_current_col,
+	     int target_column)
 {
-  if (current_column >= target_column)
-    return current_column;
+  if (my_current_col >= target_column)
+    return my_current_col;
 
   if (tabsize > 1)
     {
       int offset;
 
-      offset = tabsize - (current_column - 1) % tabsize;
-      while (current_column + offset <= target_column)
+      offset = tabsize - (my_current_col - 1) % tabsize;
+      while (my_current_col + offset <= target_column)
 	{
-	  putc (TAB, output);
-	  current_column += offset;
+	  pass_char (TAB);
+	  my_current_col += offset;
 	  offset = tabsize;
 	}
     }
 
-  while (current_column < target_column)
+  while (my_current_col < target_column)
     {
-      putc (' ', output);
-      current_column++;
+      pass_char (' ');
+      my_current_col++;
     }
 
-  return current_column;
+  return my_current_col;
 }
-
+
+static int
+preprocessor_level (struct parser_state *p)
+{
+  int level = 0;
+  while (p != 0)
+    {
+      if (p->preprocessor_indent)
+	++level;
+      p = p->next;
+    }
+  return level;
+}
+
 void
-dump_line ()
+dump_line (void)
 {				/* dump_line is the routine that actually
 				   effects the printing of the new source. It
 				   prints the label section, followed by the
@@ -226,8 +292,8 @@ dump_line ()
 				   level, followed by any comments */
   int cur_col;
   int target_col = 0;
-  static int not_first_line;
   int not_truncated = 1;
+  char *s_key;
 
   if (parser_state_tos->procname[0])
     {
@@ -242,18 +308,13 @@ dump_line ()
          rather than treat it as a normal blank line.  */
       if (parser_state_tos->use_ff)
 	{
-	  putc ('\014', output);
+	  pass_char (FORM_FEED);
 	  parser_state_tos->use_ff = false;
 	}
-      else
+      else if (!suppress_blanklines)
 	{
-	  if (suppress_blanklines > 0)
-	    suppress_blanklines--;
-	  else
-	    {
-	      parser_state_tos->bl_line = true;
-	      n_real_blanklines++;
-	    }
+	  parser_state_tos->bl_line = true;
+	  n_real_blanklines++;
 	}
     }
   else
@@ -268,55 +329,103 @@ dump_line ()
 	n_real_blanklines = 1;
 
       while (--n_real_blanklines >= 0)
-	putc (EOL, output);
+	pass_char (EOL);
       n_real_blanklines = 0;
       if (parser_state_tos->ind_level == 0)
-	parser_state_tos->ind_stmt = 0;	/* this is a class A kludge. dont do
-					   additional statement indentation
-					   if we are at bracket level 0 */
+	parser_state_tos->ind_stmt = 0;		/* This is a class A kludge. Don't do
+						   additional statement indentation
+						   if we are at bracket level 0 */
 
       if (e_lab != s_lab || e_code != s_code)
 	++code_lines;		/* keep count of lines with code */
 
       if (e_lab != s_lab)
 	{			/* print lab, if any */
-	  if (comment_open)
-	    {
-	      comment_open = 0;
-	      fprintf (output, ".*/\n");
-	    }
-	  while (e_lab > s_lab && (e_lab[-1] == ' ' || e_lab[-1] == TAB))
+	  int label_target = compute_label_target ();
+	  int pad_preproc = 0;
+	  char *skip_pound = s_lab;
+
+	  while (e_lab > s_lab && isblank (e_lab[-1]))
 	    e_lab--;
-	  cur_col = pad_output (1, compute_label_target ());
-	  if (s_lab[0] == '#' && (strncmp (s_lab, "#else", 5) == 0
-				  || strncmp (s_lab, "#endif", 6) == 0))
+
+	  cur_col = 1;
+	  if (*skip_pound == '#')
 	    {
-	      /* Treat #else and #endif as a special case because any text
-	         after #else or #endif should be converted to a comment.  */
-	      char *s = s_lab;
-	      if (e_lab[-1] == EOL)
-		e_lab--;
-	      do
-		putc (*s++, output);
-	      while (s < e_lab && 'a' <= *s && *s <= 'z');
-	      while ((*s == ' ' || *s == TAB) && s < e_lab)
-		s++;
-	      if (s < e_lab)
+	      pass_char (*skip_pound++);
+	      ++cur_col;
+
+	      for (s_key = skip_pound; s_key < e_lab && isblank (*s_key); ++s_key);
+
+	      if (s_key >= e_lab)
 		{
-		  if (s[0] == '/' && (s[1] == '*' || s[1] == '/'))
-		    fprintf (output, (tabsize > 1 ? "\t%.*s" : "  %.*s"),
-			     e_lab - s, s);
-		  else
-		    fprintf (output, (tabsize > 1
-				      ? "\t/* %.*s */"
-				      : "  /* %.*s */"),
-			     e_lab - s, s);
+		  s_key = 0;
+		}
+	      else if (preprocessor_indentation)
+		{
+		  int adj = (!strncmp (s_key, "if", (size_t) 2) ||
+			     !strncmp (s_key, "el", (size_t) 2)) ? 1 : 0;
+		  pad_preproc = ((preprocessor_level (parser_state_tos) - adj)
+				 * preprocessor_indentation);
+		  cur_col = pad_output (cur_col, cur_col + pad_preproc);
 		}
 	    }
 	  else
-	    fprintf (output, "%.*s", (int) (e_lab - s_lab), s_lab);
-	  cur_col = count_columns (cur_col, s_lab,
-				   NULL_CHAR);
+	    {
+	      s_key = 0;
+	    }
+	  cur_col = pad_output (cur_col, label_target);
+	  if (s_key != 0
+	      && (strncmp (s_key, "else", (size_t) 4) == 0
+		  || strncmp (s_key, "endif", (size_t) 5) == 0))
+	    {
+	      /* Treat #else and #endif as a special case because any text
+	         after #else or #endif should be converted to a comment.  */
+	      char *s = skip_pound;
+
+	      if (e_lab[-1] == EOL)
+		e_lab--;
+	      do
+		{
+		  /* skip 's' past the end of "endif" */
+		  pass_char (*s++);
+		}
+	      while ((s < s_key) || (s < e_lab && 'a' <= *s && *s <= 'z'));
+	      cur_col = this_column (skip_pound, s, cur_col);
+	      /* skip whitespace after "endif" */
+	      while (isblank (*s) && s < e_lab)
+		s++;
+	      if (s < e_lab)
+		{
+		  int len = (int) (e_lab - s);
+
+		  if (cur_col < else_endif_col)
+		    {
+		      (void) pad_output (cur_col, else_endif_col);
+		    }
+		  else
+		    {
+		      pass_char (' ');
+		      ++cur_col;
+		    }
+		  if (s[0] == '/' && (s[1] == '*' || s[1] == '/'))
+		    {
+		      pass_n_text (s, len);
+		    }
+		  else
+		    {
+		      pass_text ("/* ");
+		      pass_n_text (s, len);
+		      pass_text (" */");
+		      len += 2;
+		    }
+		  cur_col += len;
+		}
+	    }
+	  else
+	    {
+	      pass_n_text (skip_pound, (int) (e_lab - skip_pound));
+	      cur_col = count_columns (cur_col, skip_pound, NULL_CHAR);
+	    }
 	}
       else
 	cur_col = 1;		/* there is no label section */
@@ -324,19 +433,13 @@ dump_line ()
       parser_state_tos->pcase = false;
 
       /* Remove trailing spaces */
-      while (*(e_code - 1) == ' ' && e_code > s_code)
+      while (isblank (*(e_code - 1)) && e_code > s_code)
 	*(--e_code) = NULL_CHAR;
       if (s_code != e_code)
 	{			/* print code section, if any */
 	  char *p;
 	  int i;
 
-	  if (comment_open)
-	    {
-	      comment_open = 0;
-	      fprintf (output, ".*/\n");
-	    }
-
 	  /* If a comment begins this line, then indent it to the right
 	     column for comments, otherwise the line starts with code,
 	     so indent it for code. */
@@ -365,54 +468,58 @@ dump_line ()
 	      && s_com == e_com
 	      && buf_break > s_code && buf_break < e_code - 1)
 	    {
-	      int len;
+	      size_t len;
 
 	      for (p = s_code; p < buf_break; p++)
-		putc (*p, output);
+		pass_char (*p);
 
 	      *buf_break = '\0';
 	      cur_col = count_columns (cur_col, s_code, NULL_CHAR);
 
 	      not_truncated = 0;
-	      len = (e_code - buf_break - 1);
+	      len = (size_t) (e_code - buf_break - 1);
 	      memmove (s_code, buf_break + 1, len);
 	      e_code = s_code + len;
 	      buf_break = e_code;
-#if 0
-	      if (parser_state_tos->want_blank)
-		*e_code++ = ' ';
-#endif
 	      *e_code = '\0';
 	      break_line = 0;
 	    }
 	  else
 	    {
 	      for (p = s_code; p < e_code; p++)
-		putc (*p, output);
+		pass_char (*p);
 	      cur_col = count_columns (cur_col, s_code, NULL_CHAR);
 	    }
 	}
 
       if (s_com != e_com)
 	{
-	    {
-	      /* Here for comment printing.  This code is new as of
-	         version 1.8 */
-	      int target = parser_state_tos->com_col;
-	      char *com_st = s_com;
+	  {
+	    /* Here for comment printing.  This code is new as of
+	       version 1.8 */
+	    int target = parser_state_tos->com_col;
+	    char *com_st = s_com;
 
-	      if (cur_col > target)
-		{
-		  putc (EOL, output);
-		  cur_col = 1;
-		  ++out_lines;
-		}
-
-	      cur_col = pad_output (cur_col, target);
-	      fwrite (com_st, e_com - com_st, 1, output);
-	      cur_col += e_com - com_st;
-	      com_lines++;
-	    }
+	    if ((cur_col > target) && (*s_lab != '#'))
+	      {
+		pass_char (EOL);
+		cur_col = 1;
+		++out_lines;
+	      }
+	    else if ((cur_col == target)
+		     && (cur_col > 1)
+		     && (s_code != e_code || s_lab != e_lab))
+	      {
+		pass_char (' ');
+		++cur_col;
+	      }
+	    else
+	      {
+		(void) pad_output (cur_col, target);
+	      }
+	    pass_n_text (com_st, (int) (e_com - com_st));
+	    com_lines++;
+	  }
 	}
       else if (embedded_comment_on_line)
 	com_lines++;
@@ -420,17 +527,18 @@ dump_line ()
 
       if (parser_state_tos->use_ff)
 	{
-	  putc ('\014', output);
+	  pass_char (FORM_FEED);
 	  parser_state_tos->use_ff = false;
 	}
       else
-	putc (EOL, output);
+	pass_char (EOL);
 
       ++out_lines;
       if (parser_state_tos->just_saw_decl == 1
 	  && blanklines_after_declarations)
 	{
-	  prefix_blankline_requested = 1;
+	  if (!parser_state_tos->in_comment)
+	    prefix_blankline_requested = 1;
 	  parser_state_tos->just_saw_decl = 0;
 	}
       else
@@ -448,7 +556,7 @@ dump_line ()
 				& ~parser_state_tos->in_decl);
 
   parser_state_tos->dumped_decl_indent = 0;
-  *(e_lab  = s_lab) = '\0';	/* reset buffers */
+  *(e_lab = s_lab) = '\0';	/* reset buffers */
   if (not_truncated)
     {
       *(e_code = s_code) = '\0';
@@ -457,24 +565,27 @@ dump_line ()
   break_line = 0;
   buf_break = NULL;
 
-  *(e_com  = s_com) = '\0';
+  *(e_com = s_com) = '\0';
   parser_state_tos->ind_level = parser_state_tos->i_l_follow;
   parser_state_tos->paren_level = parser_state_tos->p_l_follow;
   if (parser_state_tos->paren_level > 0)
     {
       /* If we broke the line and the following line will
-	 begin with a rparen, the indentation is set for
-	 the column of the rparen *before* the break - reset
-	 the column to the position after the break. */
-      if (! not_truncated && *s_code == '('
+         begin with a rparen, the indentation is set for
+         the column of the rparen *before* the break - reset
+         the column to the position after the break. */
+      if (!not_truncated && *s_code == '('
 	  && parser_state_tos->paren_level >= 2)
 	{
-	  paren_target = -parser_state_tos->paren_indents[parser_state_tos->paren_level - 2];
+	  paren_target =
+	    -parser_state_tos->paren_indents[parser_state_tos->paren_level - 2];
 	  parser_state_tos->paren_indents[parser_state_tos->paren_level - 1]
-	    = parser_state_tos->paren_indents[parser_state_tos->paren_level - 2] - 1;
+	    = parser_state_tos->paren_indents[parser_state_tos->paren_level
+					      - 2] - 1;
 	}
       else
-	paren_target = -parser_state_tos->paren_indents[parser_state_tos->paren_level - 1];
+	paren_target =
+	  -parser_state_tos->paren_indents[parser_state_tos->paren_level - 1];
     }
   else
     paren_target = 0;
@@ -486,40 +597,24 @@ dump_line ()
 /* Return the column in which we should place the code about to be output. */
 
 int
-compute_code_target ()
+compute_code_target (void)
 {
   int target_col = parser_state_tos->ind_level + 1;
 
-  if (! parser_state_tos->paren_level)
+  if (!parser_state_tos->paren_level)
     {
       if (parser_state_tos->ind_stmt)
 	target_col += continuation_indent;
       return target_col;
     }
 
-  if (! lineup_to_parens)
+  if (!lineup_to_parens)
     return target_col + (continuation_indent * parser_state_tos->paren_level);
-
-#if 0
-  t = paren_target;
-  if ((w = count_columns (t, s_code, NULL_CHAR) - max_col) > 0
-      && count_columns (target_col, s_code, NULL_CHAR) <= max_col)
-    {
-      t -= w + 1;
-      if (t > target_col)
-	target_col = t;
-    }
-  else
-    target_col = t;
-
-  return target_col;
-#endif
-
   return paren_target;
 }
 
 int
-compute_label_target ()
+compute_label_target (void)
 {
   return
     (parser_state_tos->pcase
@@ -540,16 +635,13 @@ compute_label_target ()
 static struct file_buffer fileptr;
 
 struct file_buffer *
-read_file (filename)
-     char *filename;
+read_file (const char *filename)
 {
   int fd;
-  /* Required for MSDOS, in order to read files larger than 32767
-     bytes in a 16-bit world... */
-  unsigned int size;
+  long size;
 
   struct stat file_stats;
-  int namelen = strlen (filename);
+  size_t namelen = strlen (filename);
 
   fd = open (filename, O_RDONLY, 0777);
   if (fd < 0)
@@ -559,40 +651,31 @@ read_file (filename)
     fatal ("Can't stat input file %s", filename);
 
   if (file_stats.st_size == 0)
-    ERROR ("Warning: Zero-length file %s", filename, 0);
+    message (-1, "Warning: Zero-length file %s", filename);
 
-#ifdef __MSDOS__
-  if (file_stats.st_size < 0 || file_stats.st_size > (0xffff - 1))
-    fatal ("File %s is too big to read", filename);
-#else
   if (file_stats.st_size < 0)
     fatal ("System problem reading file %s", filename);
-#endif
-  fileptr.size = file_stats.st_size;
+  fileptr.size = (unsigned long) file_stats.st_size;
   if (fileptr.data != 0)
     fileptr.data = (char *) xrealloc (fileptr.data,
-				      (unsigned) file_stats.st_size + 1);
+				      (size_t) file_stats.st_size + 1);
   else
-    fileptr.data = (char *) xmalloc ((unsigned) file_stats.st_size + 1);
+    fileptr.data = (char *) xmalloc ((size_t) file_stats.st_size + 1);
 
   size = SYS_READ (fd, fileptr.data, fileptr.size);
-#ifdef __MSDOS__
-  if (size == 0xffff) /* -1 = 0xffff in 16-bit world */
-#else
   if (size < 0)
-#endif
     fatal ("Error reading input file %s", filename);
   if (close (fd) < 0)
-    fatal ("Error closeing input file %s", filename);
+    fatal ("Error closing input file %s", filename);
 
   /* Apparently, the DOS stores files using CR-LF for newlines, but
      then the DOS `read' changes them into '\n'.  Thus, the size of the
      file on disc is larger than what is read into memory.  Thanks, Bill. */
-  if (size < fileptr.size)
-    fileptr.size = size;
+  if ((size_t) size < fileptr.size)
+    fileptr.size = (size_t) size;
 
   if (fileptr.name != 0)
-    fileptr.name = (char *) xrealloc (fileptr.name, (unsigned) namelen + 1);
+    fileptr.name = (char *) xrealloc (fileptr.name, namelen + 1);
   else
     fileptr.name = (char *) xmalloc (namelen + 1);
   (void) memcpy (fileptr.name, filename, namelen);
@@ -603,21 +686,17 @@ read_file (filename)
   return &fileptr;
 }
 
-/* This should come from stdio.h and be some system-optimal number */
-#ifndef BUFSIZ
-#define BUFSIZ 1024
-#endif
-
 /* Suck the standard input into a file_buffer structure, and
    return a pointer to that structure. */
 
 struct file_buffer stdinptr;
 
 struct file_buffer *
-read_stdin ()
+read_stdin (void)
 {
-  unsigned int size = 15 * BUFSIZ;
-  int ch;
+  /* This should be some system-optimal number */
+  size_t size = 15 * BUFSIZ;
+  int ch = EOF;
   char *p;
 
   if (stdinptr.data != 0)
@@ -634,28 +713,34 @@ read_stdin ()
 	  if (ch == EOF)
 	    break;
 
-	  *p++ = ch;
+	  *p++ = (char) ch;
 	  stdinptr.size++;
 	}
 
       if (ch != EOF)
 	{
 	  size += (2 * BUFSIZ);
-	  stdinptr.data = xrealloc (stdinptr.data, (unsigned) size);
+	  stdinptr.data = xrealloc (stdinptr.data, size);
 	  p = stdinptr.data + stdinptr.size;
 	}
     }
   while (ch != EOF);
 
-  stdinptr.name = "Standard Input";
+  stdinptr.name = xstrdup ("Standard Input");
 
   stdinptr.data[stdinptr.size] = '\0';
 
   return &stdinptr;
 }
 
-/* Points to the current input buffer */
-extern struct file_buffer *current_input;
+static void
+unterminated_inhibit (int start_no)
+{
+  int save_line_no = in_line_no;
+  in_line_no = start_no + 1;
+  message (1, "unterminated INDENT-OFF comment", 0);
+  in_line_no = save_line_no;
+}
 
 /* Advance `buf_ptr' so that it points to the next line of input.
 
@@ -672,7 +757,7 @@ extern struct file_buffer *current_input;
    not recognize such lines. */
 
 void
-fill_buffer ()
+fill_buffer (void)
 {
   char *p;
   int finished_a_line;
@@ -687,8 +772,10 @@ fill_buffer ()
       bp_save = be_save = 0;
 
       /* only return if there is really something in this buffer */
-      if (buf_ptr < buf_end)
-	return;
+      if (!at_buffer_end (buf_ptr))
+	{
+	  return;
+	}
     }
 
   if (*in_prog_pos == '\0')
@@ -705,83 +792,167 @@ fill_buffer ()
   finished_a_line = 0;
   do
     {
-      while (*p == ' ' || *p == TAB)
+      /*
+       * Anything after the second "%%" line is passed to the formatter.
+       * Before the second "%%" line, "%{" and "%}" bracket lines that are
+       * passed to the formatter.  None of those control lines are passed to
+       * the formatter.
+       */
+      if (lex_or_yacc)
+	{
+	  int pass_lexcode = (lex_section >= 2);
+	  int skip_line = 0;
+
+	  if (!pass_lexcode)
+	    {
+	      if (!strncmp (p, "%%", (size_t) 2))
+		{
+		  ++lex_section;
+		  if (lex_section >= 0)
+		    {
+		      reset_output ();
+		      skip_line = (lex_section >= 2);
+		    }
+		}
+	      else if (!strncmp (p, "%{", (size_t) 2))
+		{
+		  next_lexcode = 1;
+		  skip_line = 1;
+		}
+	      else if (!strncmp (p, "%}", (size_t) 2))
+		{
+		  if (next_lexcode)
+		    {
+		      if (next_lexcode < 0)
+			{
+			  dump_line ();
+			}
+		      pass_char ('\n');
+		      next_lexcode = 0;
+		    }
+		}
+	      else if (next_lexcode)
+		{
+		  if (next_lexcode > 0)
+		    {
+		      next_lexcode = -1;
+		      reset_output ();
+		    }
+		  pass_lexcode = 1;
+		}
+	    }
+
+	  if (!pass_lexcode)
+	    {
+	      p = pass_line (p);
+	      if (skip_line)
+		{
+		  pass_char ('\n');
+		}
+	      if (!p)
+		return;
+	      continue;
+	    }
+	}
+
+      while (isblank (*p))
 	p++;
 
       /* If we are looking at the beginning of a comment, see
-	 if it turns off formatting with off-on directives. */
+         if it turns off formatting with off-on directives. */
       if (*p == '/' && (*(p + 1) == '*' || *(p + 1) == '/'))
 	{
 	  p += 2;
-	  while (*p == ' ' || *p == TAB)
+	  while (isblank (*p))
 	    p++;
 
 	  /* Skip all lines between the indent off and on directives. */
-	  if (! strncmp (p, "*INDENT-OFF*", 12))
+	  if (!strncmp (p, "*INDENT-OFF*", (size_t) 12))
 	    {
 	      char *q = cur_line;
 	      int inhibited = 1;
+	      int starting_no = in_line_no;
 
 	      if (s_com != e_com || s_lab != e_lab || s_code != e_code)
 		dump_line ();
 	      while (q < p)
-		putc (*q++, output);
+		pass_char (*q++);
 
 	      do
 		{
 		  while (*p != '\0' && *p != EOL)
-		    putc (*p++, output);
+		    pass_char (*p++);
 		  if (*p == '\0'
 		      && ((unsigned int) (p - current_input->data) == current_input->size))
 		    {
 		      buf_ptr = buf_end = in_prog_pos = p;
-		      had_eof = 1;
+		      had_eof = true;
+		      unterminated_inhibit (starting_no);
 		      return;
 		    }
 
 		  if (*p == EOL)
-		    cur_line = p + 1;
-		  putc (*p++, output);
-		  while (*p == ' ' || *p == TAB)
-		    putc (*p++, output);
+		    {
+		      ++in_line_no;
+		      cur_line = p + 1;
+		    }
+		  pass_char (*p++);
+		  while (isblank (*p))
+		    pass_char (*p++);
 
 		  if (*p == '/' && (*(p + 1) == '*' || *(p + 1) == '/'))
 		    {
-		      /* We've hit a comment.  See if turns formatting
-			 back on. */
-		      putc (*p++, output);
-		      putc (*p++, output);
-		      while (*p == ' ' || *p == TAB)
-			putc (*p++, output);
-		      if (! strncmp (p, "*INDENT-ON*", 11))
+		      /* We've hit a comment.  See if it turns formatting
+		         back on. */
+		      pass_char (*p++);
+		      pass_char (*p++);
+		      while (isblank (*p))
+			pass_char (*p++);
+		      if (!strncmp (p, "*INDENT-ON*", (size_t) 11))
 			{
 			  do
 			    {
 			      while (*p != '\0' && *p != EOL)
-				putc (*p++, output);
+				pass_char (*p++);
 			      if (*p == '\0'
-				  && ((unsigned int) (p - current_input->data) == current_input->size))
+				  && ((unsigned int) (p -
+						      current_input->data)
+				      == current_input->size))
 				{
 				  buf_ptr = buf_end = in_prog_pos = p;
-				  had_eof = 1;
+				  had_eof = true;
 				  return;
 				}
 			      else
 				{
 				  if (*p == EOL)
 				    {
+				      ++in_line_no;
 				      inhibited = 0;
 				      cur_line = p + 1;
 				    }
-				  putc (*p++, output);
+				  pass_char (*p++);
 				}
 			    }
 			  while (inhibited);
 			}
+		      else if (!strncmp (p, "*INDENT-OFF*", (size_t) 12))
+			{
+			  unterminated_inhibit (starting_no);
+			  starting_no = in_line_no;
+			}
+		      else if (!strncmp (p, "*INDENT-", (size_t) 8))
+			{
+			  message (1, "unexpected INDENT-comment", 0);
+			}
 		    }
 		}
 	      while (inhibited);
 	    }
+	  else if (!strncmp (p, "*INDENT-EQLS*", (size_t) 13))
+	    {
+	      indent_eqls = -1;
+	    }
 	}
 
       while (*p != '\0' && *p != EOL)
@@ -790,13 +961,14 @@ fill_buffer ()
       /* Here for newline -- finish up unless formatting is off */
       if (*p == EOL)
 	{
+	  ++in_line_no;
 	  finished_a_line = 1;
 	  in_prog_pos = p + 1;
 	}
       /* Here for embedded NULLs */
       else if ((unsigned int) (p - current_input->data) < current_input->size)
 	{
-	  WARNING ("Warning: File %s contains NULL-characters\n",
+	  message (-1, "Warning: File %s contains NULL-characters\n",
 		   current_input->name, 0);
 	  p++;
 	}
@@ -804,7 +976,7 @@ fill_buffer ()
       else
 	{
 	  in_prog_pos = p;
-	  finished_a_line = 1;
+	  break;
 	}
     }
   while (!finished_a_line);
@@ -812,113 +984,35 @@ fill_buffer ()
   buf_ptr = cur_line;
   buf_end = in_prog_pos;
   buf_break = NULL;
-}
-
 
-INLINE void
-writefdef (f, nm)
-     struct fstate *f;
-     unsigned int nm;
-{
-  fprintf (output, ".ds f%c %s\n.nr s%c %d\n",
-	   (int) nm, f->font, nm, (int) f->size);
-}
+  /*
+   * Turn off the indent-eqls feature on the next blank line.
+   */
+  if (indent_eqls > 0)
+    {
+      int blank = true;
 
-#if 0
-/* Write characters starting at S to change the font from OF to NF.  Return a
-   pointer to the character after the last character written. For troff mode
-   only.  */
-char *
-chfont (of, nf, s)
-     struct fstate *of, *nf;
-     char *s;
-{
-  if (of->font[0] != nf->font[0]
-      || of->font[1] != nf->font[1])
-    {
-      *s++ = '\\';
-      *s++ = 'f';
-      if (nf->font[1])
+      for (p = buf_ptr; p < buf_end; ++p)
 	{
-	  *s++ = '(';
-	  *s++ = nf->font[0];
-	  *s++ = nf->font[1];
+	  if (!isspace (UChar (*p)))
+	    {
+	      blank = false;
+	      break;
+	    }
 	}
-      else
-	*s++ = nf->font[0];
-    }
-  if (nf->size != of->size)
-    {
-      *s++ = '\\';
-      *s++ = 's';
-      if (nf->size < of->size)
+      if (blank)
 	{
-	  *s++ = '-';
-	  *s++ = '0' + of->size - nf->size;
-	}
-      else
-	{
-	  *s++ = '+';
-	  *s++ = '0' + nf->size - of->size;
+	  indent_eqls = 0;
+	  indent_eqls_1st = 0;
 	}
     }
-  return s;
-}
 
-void
-parsefont (f, s0)
-     struct fstate *f;
-     char *s0;
-{
-  char *s = s0;
-  int sizedelta = 0;
-  int i;
-
-  f->size = 0;
-  f->allcaps = 1;
-  for (i = 0; i < 4; i++)
-    f->font[i] = 0;
-
-  while (*s)
+  if (debug)
     {
-      if (isdigit (*s))
-	f->size = f->size * 10 + *s - '0';
-      else if (isupper (*s))
-	if (f->font[0])
-	  f->font[1] = *s;
-	else
-	  f->font[0] = *s;
-      else if (*s == 'c')
-	f->allcaps = 1;
-      else if (*s == '+')
-	sizedelta++;
-      else if (*s == '-')
-	sizedelta--;
-      else
-	{
-	  fprintf (stderr, "indent: bad font specification: %s\n", s0);
-	  exit (invocation_error);
-	}
-      s++;
+      printf ("%6d: %s%.*s%s",
+	      in_line_no,
+	      finished_a_line ? "" : "*",
+	      (int) (buf_end - buf_ptr), buf_ptr,
+	      finished_a_line ? "" : "\n");
     }
-  if (f->font[0] == 0)
-    f->font[0] = 'R';
-  if (bodyf.size == 0)
-    bodyf.size = 11;
-  if (f->size == 0)
-    f->size = bodyf.size + sizedelta;
-  else if (sizedelta > 0)
-    f->size += bodyf.size;
-  else
-    f->size = bodyf.size - f->size;
 }
-#endif
-
-#ifdef DEBUG
-void
-dump_debug_line ()
-{
-  fprintf (output, "\n*** Debug output marker line ***\n");
-}
-
-#endif
