@@ -1,4 +1,7 @@
-/* Copyright (c) 1994, Joseph Arceneaux.  All rights reserved
+/*
+   Copyright 1999-2002,2010, Thomas E. Dickey
+
+   Copyright (c) 1994, Joseph Arceneaux.  All rights reserved
 
    Copyright (c) 1985 Sun Microsystems, Inc. Copyright (c) 1980 The Regents
    of the University of California. Copyright (c) 1976 Board of Trustees of
@@ -21,14 +24,141 @@
 
 struct parser_state *parser_state_tos;
 
-#define INITIAL_BUFFER_SIZE 1000
+#define INITIAL_BUFFER_SIZE ((size_t) 1000)
 #define INITIAL_STACK_SIZE 2
 
+const char *
+parsecode2s (enum codes value)
+{
+  const char *result = "?";
+  switch (value)
+    {
+    case code_eof:
+      result = "EOF";
+      break;
+    case newline:
+      result = "newline";
+      break;
+    case lparen:
+      result = "left parenthesis";
+      break;
+    case rparen:
+      result = "right parenthesis";
+      break;
+    case start_token:
+      result = "start_token";
+      break;
+    case unary_op:
+      result = "unary_op";
+      break;
+    case binary_op:
+      result = "binary_op";
+      break;
+    case postop:
+      result = "postop";
+      break;
+    case question:
+      result = "question mark";
+      break;
+    case casestmt:
+      result = "casestmt";
+      break;
+    case colon:
+      result = "colon";
+      break;
+    case doublecolon:		/* For C++ class methods. */
+      result = "doublecolon";
+      break;
+    case semicolon:
+      result = "semicolon";
+      break;
+    case lbrace:
+      result = "left brace";
+      break;
+    case rbrace:
+      result = "right brace";
+      break;
+    case ident:
+      result = "identifier or string";
+      break;
+    case overloaded:		/* For C++ overloaded operators (like +) */
+      result = "overloaded";
+      break;
+    case cpp_operator:
+      result = "cpp_operator";
+      break;
+    case comma:
+      result = "comma";
+      break;
+    case comment:		/* A "slash-star" comment */
+      result = "comment";
+      break;
+    case cplus_comment:	/* A C++ "slash-slash" */
+      result = "cplus_comment";
+      break;
+    case swstmt:
+      result = "switch-statement";
+      break;
+    case preesc:		/* '#'.  */
+      result = "preesc";
+      break;
+    case form_feed:
+      result = "form_feed";
+      break;
+    case decl:
+      result = "declaration";
+      break;
+    case sp_paren:		/* if, for, or while token */
+      result = "sp_paren";
+      break;
+    case sp_nparen:
+      result = "sp_nparen";
+      break;
+    case ifstmt:
+      result = "if-statement";
+      break;
+    case whilestmt:
+      result = "while-statement";
+      break;
+    case forstmt:
+      result = "for-statement";
+      break;
+    case stmt:
+      result = "statement";
+      break;
+    case stmtl:
+      result = "statement list";
+      break;
+    case elselit:
+      result = "elselit";
+      break;
+    case dolit:
+      result = "dolit";
+      break;
+    case dohead:
+      result = "do-head";
+      break;
+    case dostmt:
+      result = "do-statement";
+      break;
+    case ifhead:
+      result = "if-head";
+      break;
+    case elsehead:
+      result = "else-head";
+      break;
+    case struct_delim:		/* '.' or "->" */
+      result = "struct_delim";
+      break;
+    }
+  return result;
+}
+
 void
-init_parser ()
+init_parser (void)
 {
   parser_state_tos
-  = (struct parser_state *) xmalloc (sizeof (struct parser_state));
+    = (struct parser_state *) xmalloc (sizeof (struct parser_state));
 
   parser_state_tos->p_stack_size = INITIAL_STACK_SIZE;
   parser_state_tos->p_stack
@@ -37,7 +167,7 @@ init_parser ()
     = (int *) xmalloc (INITIAL_STACK_SIZE * sizeof (int));
   parser_state_tos->cstk
     = (int *) xmalloc (INITIAL_STACK_SIZE * sizeof (int));
-  parser_state_tos->paren_indents = (short *) xmalloc (sizeof (short));
+  parser_state_tos->paren_indents = (int *) xmalloc (sizeof (int));
 
   /* Although these are supposed to grow if we reach the end,
      I can find no place in the code which does this. */
@@ -47,7 +177,8 @@ init_parser ()
 
   save_com.size = INITIAL_BUFFER_SIZE;
   save_com.end = save_com.ptr = xmalloc (save_com.size);
-  save_com.len = save_com.column = 0;
+  save_com.len = 0;
+  save_com.column = 0;
 
   di_stack_alloc = 2;
   di_stack = (int *) xmalloc (di_stack_alloc * sizeof (*di_stack));
@@ -55,62 +186,66 @@ init_parser ()
 }
 
 void
-reset_parser ()
+reset_parser (void)
 {
-  parser_state_tos->next = 0;
-  parser_state_tos->tos = 0;
+  /* *INDENT-EQLS* */
+  parser_state_tos->next               = 0;
+  parser_state_tos->tos                = 0;
   parser_state_tos->paren_indents_size = 1;
-  parser_state_tos->p_stack[0] = stmt;	/* this is the parser's stack */
-  parser_state_tos->last_nl = true;	/* this is true if the last thing
-					   scanned was a newline */
-  parser_state_tos->last_token = start_token;
-  parser_state_tos->box_com = false;
-  parser_state_tos->cast_mask = 0;
-  parser_state_tos->noncast_mask = 0;
-  parser_state_tos->sizeof_mask = 0;
-  parser_state_tos->block_init = false;
-  parser_state_tos->block_init_level = 0;
-  parser_state_tos->col_1 = false;
-  parser_state_tos->com_col = 0;
-  parser_state_tos->dec_nest = 0;
-  parser_state_tos->i_l_follow = 0;
-  parser_state_tos->ind_level = 0;
-  parser_state_tos->last_u_d = false;
-  parser_state_tos->p_l_follow = 0;
-  parser_state_tos->paren_level = 0;
-  parser_state_tos->paren_depth = 0;
-  parser_state_tos->search_brace = false;
-  parser_state_tos->use_ff = false;
-  parser_state_tos->its_a_keyword = false;
-  parser_state_tos->sizeof_keyword = false;
+  parser_state_tos->p_stack[0]         = stmt;
+  parser_state_tos->last_nl            = true;
+  parser_state_tos->last_token         = start_token;
+  parser_state_tos->box_com            = false;
+  parser_state_tos->cast_mask          = 0;
+  parser_state_tos->noncast_mask       = 0;
+  parser_state_tos->sizeof_mask        = 0;
+  parser_state_tos->block_init         = false;
+  parser_state_tos->block_init_level   = 0;
+  parser_state_tos->col_1              = false;
+  parser_state_tos->com_col            = 0;
+  parser_state_tos->dec_nest           = 0;
+  parser_state_tos->i_l_follow         = 0;
+  parser_state_tos->ind_level          = 0;
+  parser_state_tos->inner_stmt         = 0;
+  parser_state_tos->last_u_d           = false;
+  parser_state_tos->p_l_follow         = 0;
+  parser_state_tos->paren_level        = 0;
+  parser_state_tos->paren_depth        = 0;
+  parser_state_tos->search_brace       = false;
+  parser_state_tos->use_ff             = false;
+  parser_state_tos->its_a_keyword      = false;
+  parser_state_tos->sizeof_keyword     = false;
   parser_state_tos->dumped_decl_indent = false;
   parser_state_tos->in_parameter_declaration = false;
-  parser_state_tos->just_saw_decl = false;
-  parser_state_tos->in_decl = false;
-  parser_state_tos->decl_on_line = false;
-  parser_state_tos->in_or_st = false;
-  parser_state_tos->bl_line = true;
-  parser_state_tos->want_blank = false;
-  parser_state_tos->in_stmt = false;
-  parser_state_tos->ind_stmt = false;
-  parser_state_tos->procname = "\0";
-  parser_state_tos->procname_end = "\0";
-  parser_state_tos->classname = "\0";
-  parser_state_tos->classname_end = "\0";
-  parser_state_tos->pcase = false;
-  parser_state_tos->dec_nest = 0;
+  parser_state_tos->in_comment         = false;
+  parser_state_tos->just_saw_decl      = false;
+  parser_state_tos->in_decl            = false;
+  parser_state_tos->decl_on_line       = false;
+  parser_state_tos->in_or_st           = false;
+  parser_state_tos->bl_line            = true;
+  parser_state_tos->want_blank         = false;
+  parser_state_tos->in_stmt            = false;
+  parser_state_tos->ind_stmt           = false;
+  parser_state_tos->procname           = "\0";
+  parser_state_tos->procname_end       = "\0";
+  parser_state_tos->classname          = "\0";
+  parser_state_tos->classname_end      = "\0";
+  parser_state_tos->pcase              = false;
+  parser_state_tos->dec_nest           = 0;
+  parser_state_tos->preprocessor_indent = false;
+  parser_state_tos->il[0]              = 0;
+  parser_state_tos->cstk[0]            = 0;
+  parser_state_tos->paren_indents[0]   = 0;
 
-  parser_state_tos->il[0] = 0;
-  parser_state_tos->cstk[0] = 0;
-  parser_state_tos->paren_indents[0] = 0;
-
-  save_com.len = save_com.column = 0;
+  save_com.len = 0;
+  save_com.column = 0;
 
   di_stack[parser_state_tos->dec_nest] = 0;
 
-  l_com = combuf + INITIAL_BUFFER_SIZE - 5;
-  l_lab = labbuf + INITIAL_BUFFER_SIZE - 5;
-  l_code = codebuf + INITIAL_BUFFER_SIZE - 5;
+  /* *INDENT-EQLS* */
+  l_com     = combuf + INITIAL_BUFFER_SIZE - 5;
+  l_lab     = labbuf + INITIAL_BUFFER_SIZE - 5;
+  l_code    = codebuf + INITIAL_BUFFER_SIZE - 5;
   combuf[0] = codebuf[0] = labbuf[0] = ' ';
   combuf[1] = codebuf[1] = labbuf[1] = '\0';
 
@@ -120,22 +255,27 @@ reset_parser ()
   s_code = e_code = codebuf + 1;
   s_com = e_com = combuf + 1;
 
-  line_no = 1;
+  in_line_no = 0;
+  out_line_no = 1;
   had_eof = false;
   break_comma = false;
   bp_save = 0;
   be_save = 0;
+
   if (tabsize <= 0)
     tabsize = 1;
+
+  indent_eqls = 0;
+  indent_eqls_1st = 0;
 }
 
 /* like ++parser_state_tos->tos but checks for stack overflow and extends
    stack if necessary.  */
 
 int
-inc_pstack ()
+inc_pstack (void)
 {
-  if (++parser_state_tos->tos >= parser_state_tos->p_stack_size)
+  if (++parser_state_tos->tos >= (int) parser_state_tos->p_stack_size)
     {
       parser_state_tos->p_stack_size *= 2;
       parser_state_tos->p_stack = (enum codes *)
@@ -154,71 +294,48 @@ inc_pstack ()
   return parser_state_tos->tos;
 }
 
-#ifdef DEBUG
-static char **debug_symbol_strings;
+#define SHOW_PARSER(name) \
+  if (s_##name < e_##name) \
+    printf(#name ":%.*s\n", (int) (e_##name - s_##name), s_##name)
 
 void
-debug_init ()
+show_parser (const char *fn, int ln)
 {
-  int size = ((int) period + 4) * sizeof (char *);
-
-  debug_symbol_strings = (char **) xmalloc (size);
-
-  debug_symbol_strings[code_eof] = "code_eof";
-  debug_symbol_strings[newline] = "newline";
-  debug_symbol_strings[lparen] = "lparen";
-  debug_symbol_strings[rparen] = "rparen";
-  debug_symbol_strings[unary_op] = "unary_op";
-  debug_symbol_strings[binary_op] = "binary_op";
-  debug_symbol_strings[postop] = "postop";
-  debug_symbol_strings[question] = "question";
-  debug_symbol_strings[casestmt] = "casestmt";
-  debug_symbol_strings[colon] = "colon";
-  debug_symbol_strings[semicolon] = "semicolon";
-  debug_symbol_strings[lbrace] = "lbrace";
-  debug_symbol_strings[rbrace] = "rbrace";
-  debug_symbol_strings[ident] = "ident";
-  debug_symbol_strings[comma] = "comma";
-  debug_symbol_strings[comment] = "comment";
-  debug_symbol_strings[swstmt] = "swstmt";
-  debug_symbol_strings[preesc] = "preesc";
-  debug_symbol_strings[form_feed] = "form_feed";
-  debug_symbol_strings[decl] = "decl";
-  debug_symbol_strings[sp_paren] = "sp_paren";
-  debug_symbol_strings[sp_nparen] = "sp_nparen";
-  debug_symbol_strings[ifstmt] = "ifstmt";
-  debug_symbol_strings[whilestmt] = "whilestmt";
-  debug_symbol_strings[forstmt] = "forstmt";
-  debug_symbol_strings[stmt] = "stmt";
-  debug_symbol_strings[stmtl] = "stmtl";
-  debug_symbol_strings[elselit] = "elselit";
-  debug_symbol_strings[dolit] = "dolit";
-  debug_symbol_strings[dohead] = "dohead";
-  debug_symbol_strings[dostmt] = "dostmt";
-  debug_symbol_strings[ifhead] = "ifhead";
-  debug_symbol_strings[elsehead] = "elsehead";
-  debug_symbol_strings[period] = "period";
+  printf ("<<%s@%d>>\n", fn, ln);
+  SHOW_PARSER (lab);
+  SHOW_PARSER (code);
+  SHOW_PARSER (com);
 }
 
-#endif
-
-enum exit_values
-parse (tk)
-     enum codes tk;		/* the code for the construct scanned */
+void
+show_pstack (void)
 {
   int i;
-  enum exit_values my_value = total_success;
 
-#ifdef DEBUG
   if (debug)
     {
-      if (tk >= code_eof && tk <= period)
-	printf ("Parse: %s\n", debug_symbol_strings[tk]);
-      else
-	printf ("Parse: Unknown code: %d for %s\n",
-		(int) tk, token ? token : "NULL");
+      printf ("\nParseStack [%d]:\n", (int) parser_state_tos->p_stack_size);
+      for (i = 1; i <= parser_state_tos->tos; ++i)
+	printf ("  [%3d] =>  indent:%3d  stack: %2d (%s)\n",
+		(int) i,
+		(int) parser_state_tos->il[i],
+		(int) parser_state_tos->p_stack[i],
+		parsecode2s (parser_state_tos->p_stack[i]));
+      printf ("\n");
+    }
+}
+
+enum exit_values
+parse (enum codes tk)		/* the code for the construct scanned */
+{
+  int i;
+
+  if (debug)
+    {
+      if (debug > 1)
+	printf ("%s:%d:%d: ", in_name, in_line_no, token_col ());
+      printf ("Parse: %s\n", parsecode2s (tk));
     }
-#endif
 
   while (parser_state_tos->p_stack[parser_state_tos->tos] == ifhead
 	 && tk != elselit)
@@ -265,18 +382,20 @@ parse (tk)
 	  && else_if)		/* "else if ..." */
 	parser_state_tos->i_l_follow
 	  = parser_state_tos->il[parser_state_tos->tos];
+      /* FALLTHRU */
     case dolit:		/* 'do' */
+      /* FALLTHRU */
     case forstmt:		/* for (...) */
       inc_pstack ();
       parser_state_tos->p_stack[parser_state_tos->tos] = tk;
       parser_state_tos->il[parser_state_tos->tos]
 	= parser_state_tos->ind_level = parser_state_tos->i_l_follow;
-      parser_state_tos->i_l_follow += ind_size;	/* subsequent statements
-						   should be indented 1 */
+      parser_state_tos->i_l_follow += ind_size;		/* subsequent statements
+							   should be indented 1 */
       parser_state_tos->search_brace = btype_2;
       break;
 
-    case lbrace:		/* scanned { */
+    case lbrace:		/* scanned left curly-brace */
       break_comma = false;	/* don't break comma in an initial list */
       if (parser_state_tos->p_stack[parser_state_tos->tos] == stmt
 	  || parser_state_tos->p_stack[parser_state_tos->tos] == stmtl)
@@ -304,17 +423,15 @@ parse (tk)
 	      /* it is a group as part of a while, for, etc. */
 
 	      /* For -bl formatting, indent by brace_indent additional spaces
-	         e.g. if (foo == bar) { <--> brace_indent spaces (in this
-	         example, 4) */
+	       * e.g.,
+	       *        if (foo == bar) { <--> brace_indent spaces
+	       *        }
+	       * (in this example, 4)
+	       */
 	      if (!btype_2)
 		{
 		  parser_state_tos->ind_level += brace_indent;
 		  parser_state_tos->i_l_follow += brace_indent;
-#if 0
-		  if (parser_state_tos->p_stack[parser_state_tos->tos]
-		      == swstmt)
-		    case_ind += brace_indent;
-#endif
 		}
 
 	      if (parser_state_tos->p_stack[parser_state_tos->tos] == swstmt
@@ -359,8 +476,9 @@ parse (tk)
 
       if (parser_state_tos->p_stack[parser_state_tos->tos] != ifhead)
 	{
-	  ERROR ("Unmatched 'else'", 0, 0);
-	  my_value = indent_error;
+	  message (0, "Unmatched 'else' (stack has %s rather than %s)",
+		   parsecode2s (parser_state_tos->p_stack[parser_state_tos->tos]),
+		   parsecode2s (ifhead));
 	}
       else
 	{
@@ -377,7 +495,7 @@ parse (tk)
 	}
       break;
 
-    case rbrace:		/* scanned a } */
+    case rbrace:		/* scanned a right curly-brace */
       /* stack should have <lbrace> <stmt> or <lbrace> <stmtl> */
       if (parser_state_tos->p_stack[parser_state_tos->tos - 1] == lbrace)
 	{
@@ -385,34 +503,36 @@ parse (tk)
 	    = parser_state_tos->il[--parser_state_tos->tos];
 	  parser_state_tos->p_stack[parser_state_tos->tos] = stmt;
 	}
+      else if (parser_state_tos->p_stack[parser_state_tos->tos - 1] == stmtl)
+	{
+	  message (token_col (), "Did not expect a statement here");
+	  parser_state_tos->ind_level = parser_state_tos->i_l_follow
+	    = parser_state_tos->il[--parser_state_tos->tos];
+	}
       else
 	{
-	  ERROR ("Stmt nesting error.", 0, 0);
-	  my_value = indent_error;
+	  message (0, "After %s, expected %s on stack, not %s",
+		   parsecode2s (rbrace),
+		   parsecode2s (lbrace),
+		   parsecode2s
+		   (parser_state_tos->p_stack[parser_state_tos->tos - 1]));
 	}
       break;
 
     case swstmt:		/* had switch (...) */
       inc_pstack ();
       parser_state_tos->p_stack[parser_state_tos->tos] = swstmt;
-#if 0
-      parser_state_tos->cstk[parser_state_tos->tos] = case_ind;
-#endif
       parser_state_tos->cstk[parser_state_tos->tos]
 	= case_indent + parser_state_tos->i_l_follow;
-      if (! btype_2)
+      if (!btype_2)
 	parser_state_tos->cstk[parser_state_tos->tos] += brace_indent;
 
       /* save current case indent level */
       parser_state_tos->il[parser_state_tos->tos]
 	= parser_state_tos->i_l_follow;
       /* case labels should be one level down from switch, plus
-	 `case_indent' if any.  Then, statements should be the `ind_size'
-	 further. */
-#if 0
-      case_ind = parser_state_tos->i_l_follow + case_indent;
-      parser_state_tos->i_l_follow += case_indent + ind_size;
-#endif
+         `case_indent' if any.  Then, statements should be the `ind_size'
+         further. */
       parser_state_tos->i_l_follow += ind_size;
       parser_state_tos->search_brace = btype_2;
       break;
@@ -433,24 +553,14 @@ parse (tk)
 	}
       break;
 
-      /* This is a fatal error which cases the program to exit. */
+      /* This is a fatal error which causes the program to exit. */
     default:
-      fatal ("Unknown code to parser", 0, 0);
+      fatal ("Unknown code to parser");
     }
 
   reduce ();			/* see if any reduction can be done */
 
-#ifdef DEBUG
-  if (debug)
-    {
-      printf ("\nParseStack [%d]:\n", (int) parser_state_tos->p_stack_size);
-      for (i = 1; i <= parser_state_tos->tos; ++i)
-	printf ("  stack[%d] =>   stack: %d   ind_level: %d\n",
-		(int) i, (int) parser_state_tos->p_stack[i],
-		(int) parser_state_tos->il[i]);
-      printf ("\n");
-    }
-#endif
+  show_pstack ();
 
   return total_success;
 }
@@ -462,12 +572,18 @@ FUNCTION: Implements the reduce part of the parsing algorithm
 ALGORITHM: The following reductions are done.  Reductions are repeated until
    no more are possible.
 
-Old TOS		     New TOS <stmt> <stmt>	     <stmtl> <stmtl> <stmt>
-   <stmtl> do <stmt>		     dohead <dohead> <whilestmt>
-   <dostmt> if <stmt>		     "ifstmt" switch <stmt>	     <stmt>
-   decl <stmt>		     <stmt> "ifelse" <stmt>	     <stmt> for
-   <stmt>		     <stmt> while <stmt>		     <stmt>
-   "dostmt" while	     <stmt>
+   Old TOS                   New TOS
+   <stmt> <stmt>             <stmtl>
+   <stmtl> <stmt>            <stmtl>
+   do <stmt>                 dohead
+   <dohead> <whilestmt>      <dostmt>
+   if <stmt>                 "ifstmt"
+   switch <stmt>             <stmt>
+   decl <stmt>               <stmt>
+   "ifelse" <stmt>           <stmt>
+   for <stmt>                <stmt>
+   while <stmt>              <stmt>
+   "dostmt" while            <stmt>
 
 On each reduction, parser_state_tos->i_l_follow (the indentation for the
    following line) is set to the indentation level associated with the old
@@ -492,7 +608,7 @@ HISTORY: initial coding 	November 1976	D A Willcox of CAC
 \*----------------------------------------------*/
 
 void
-reduce ()
+reduce (void)
 {
   int i;
 
@@ -535,9 +651,6 @@ reduce ()
 
 	    case swstmt:
 	      /* <switch> <stmt> */
-#if 0
-	      case_ind = parser_state_tos->cstk[parser_state_tos->tos - 1];
-#endif
 
 	    case decl:		/* finish of a declaration */
 	    case elsehead:
@@ -560,9 +673,6 @@ reduce ()
 	  if (parser_state_tos->p_stack[parser_state_tos->tos - 1] == dohead)
 	    {
 	      /* it is termination of a do while */
-#if 0
-	      parser_state_tos->p_stack[--parser_state_tos->tos] = stmt;
-#endif
 	      parser_state_tos->p_stack[--parser_state_tos->tos] = dostmt;
 	      break;
 	    }
@@ -581,8 +691,8 @@ reduce ()
    necessary to make sure that "int foo(), bar()" gets formatted correctly
    under -bc.  */
 
-INLINE void
-parse_lparen_in_decl ()
+void
+parse_lparen_in_decl (void)
 {
   inc_pstack ();
   parser_state_tos->p_stack[parser_state_tos->tos] = stmt;
